import pandas as pd
from flatten_dict import flatten
from sacred import Experiment

def exp_recorder(exp_name,args,metrics,mongo_connector):
    """
    Function which will store output of your exp
    exp_name: name of your experiment
    args: dict with args used in exp
    metrics: output of gym.workout
    mongo_connector: sacred.observers.MongoObserver object which handle connection with database
    """
    ex = Experiment(exp_name)
    ex.observers.append(mongo_connector)
    flatten_args = flatten(args,reducer=lambda a,b: '{}_{}'.format(a,b) if a else b)
    flatten_args['true_args'] = args
    ex.add_config(flatten_args)
    result = metrics[list(metrics)[0]]['best_valid_metric']

    @ex.main
    def store():
        metric_store(metrics,ex)
        return result
    ex.run()

def metric_store(results,ex,prefix=''):
    """
    helper function to iterate over nasted dict created to handle and store gym metrics and losses
    """
    print(results)
    for result in results:
        if isinstance(results[result],dict):
            metric_store(results[result],ex,'{}.{}'.format(prefix,result))
        else:
            for scalar in results[result]:
                ex.log_scalar('{}.{}'.format(prefix,result)[1:-1],scalar)
                
def exp_downloader(exp_name,mongo_connector):
    """
    exp_name: name of expermient used to store outcome in MongoDB
    mongo_connector: pymnogo.MongoClient object connected to database
    """
    db = mongo_connector.ai
    cursor = db.runs.find({'experiment.name':exp_name})
    exp_list = []
    true_list = []
    for exp in cursor:
        new_exp = exp['config']
        true_args = new_exp.pop('true_args')
        exp_list.append(new_exp)
        true_list.append(true_args)
    
    return pd.concat([pd.DataFrame(exp) for exp in exp_list]), true_list
    


import numpy as np

def conf_lr_updater(min_lr,max_lr,cycles,diff_lr=False):
    '''
    configuration of lr_updater with lr_bounds and option for differential learning rate implementation
    min_lr - minimum learning rate value
    max_lr - maximum learning rate value
    cycles - how many times learning rate should bounce from boarders during one epoch
    diff_lr - False if there is only one learning rate level and list of learning rate ratios for every part of neural net
    '''
    def lr_updater(obj,diff_lr=diff_lr):
        step_size = obj.train_len / cycles
        cycle = np.floor(1 + obj.train_iters[0] / (2 * step_size))
        x = np.abs(obj.train_iters[0] / step_size - 2 * cycle + 1)
        lr = min_lr + (max_lr - min_lr) * np.maximum(0, (1 - x))
        if not diff_lr:
            diff_lr = [1] * len(obj.optimizer.param_groups)
        for p, lr_ratio in zip(obj.optimizer.param_groups,diff_lr):
            p['lr'] = lr*lr_ratio
    return lr_updater



import os
import pandas as pd
from google.cloud import translate
from tqdm import tqdm


class GoogleTranslationClient:
    def __init__(self, cred_path=None):
        import os
        if cred_path is not None:
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = cred_path

        from google.cloud import translate
        self.client = translate.Client()

    def get_all_langs(self):
        return [lang['language'] for lang in self.client.get_languages()]

    def translate(self, texts, source_lang, target_lang):
        return [text['translatedText'] for text in \
                self.client.translate(texts, source_language=source_lang,
                                             target_language=target_lang)]

def augment_data(client, inputs, labels, langs='all', source_lang='pl', drop_duplicates=True):
    from tqdm import tqdm
    import pandas as pd
    
    source_texts = inputs
    
    if langs == 'all':
        langs = client.get_all_langs()
        
    aug_texts = []
    langs_column = [source_lang]*len(source_texts)
    n_langs = 1
    for lang in tqdm(langs):
        try:
            translated_texts = client.translate(source_texts, source_lang=source_lang, target_lang=lang)
            aug_texts += client.translate(translated_texts, source_lang=lang, target_lang=source_lang)
            langs_column += [lang]*len(translated_texts)
            n_langs += 1
        except:
            print(f'Language {lang} cannot be used.')
        
    aug_df = pd.DataFrame({'input': source_texts + aug_texts, 'label': labels*n_langs, 'lang': langs_column})
    
    if drop_duplicates:
        aug_df.drop_duplicates(subset='input', inplace=True)
        
    return aug_df.input.tolist(), aug_df.label.tolist(), aug_df.lang.tolist()


import torch
from sklearn.metrics import roc_auc_score

'''
metric classes should always have at least two static methods 
`get_best` which takes list of metrics and return best metric
`evaluate` which takes list of labels and model predictions per batch and evaluate metric
'''

class Perplexity:
    @staticmethod
    def get_best(metrics):
        return min(metrics)

    @staticmethod
    def evaluate(out,lab):
        loss = torch.cat(out)
        return (loss.sum()/loss.size(0)).exp().item()

class Accuracy:
    @staticmethod
    def get_best(metrics):
        return max(metrics)
    
    @staticmethod
    def evaluate(out,lab):
        out = torch.cat(out)
        lab = torch.cat(lab)
        out_max = out.argmax(dim=1)
        score = out_max == lab
        return score.float().mean().item()

class ROC_AUC:
    @staticmethod
    def get_best(metrics):
        return max(metrics)
    
    @staticmethod
    def evaluate(out,lab):
        out = torch.cat(out)
        lab = torch.cat(lab)
        return roc_auc_score(lab, out[:, 1])




from torch.nn.utils import clip_grad_norm_

def conf_gc(max_norm, norm_type=2):
    '''
    configurator of gradient clipping function which can easily be applied to gym
    '''
    def gc(obj):
        norm = clip_grad_norm_(obj.model.parameters(),max_norm,norm_type)
        return norm
    return gc

def conf_gradients_norm(norm_type=2):
    '''
    configurator which let you change norm_type which will be logged during training
    '''
    def gradients_norm(obj):
            total_norm = 0
            parameters = obj.model.parameters()
            parameters = list(filter(lambda p: p.grad is not None, parameters))
            for p in parameters:
                param_norm = p.grad.data.norm(norm_type)
                total_norm += param_norm.item() ** norm_type
            total_norm = total_norm ** (1. / norm_type)
            return total_norm
    return gradients_norm